==> app.js-0.0.1.tgz: package/lib/App.js <==

var path = require('path'),
    fs = require('fs'),
    url = require('url'),
    _ = require('underscore'),
    vm = require('vm'),
    async = require('async'),
    assert = require('assert').ok,
    dandy = require('dandy/errors'),
    abind = require('dandy/errors').abind,
    ibind = require('dandy/errors').ibind,
    Cache = require('diskcache').Cache,
    cacheware = require('express-cache'),
    hascan = require('hascan'),
    mime = require('mime'),
    pkg = require('./pkg');
    Renderer = require('./Renderer').Renderer;

// *************************************************************************************************

var defaults = {
    title: '',
    client: null,
    cdn: null,
    api: null,
    apiCache: null,
    cache: null,
    base: "/",
    resources: "app.js",
    language: 'en',
    charset: 'utf8',
    offline: false,
    viewport: null,
    rss: null,
    appcache: null,
    webAppCapable: false,
    statusBarStyle: null,
    inlineImageMaxSize: 1000,
    htmlHeader: '<!-- To view the original source of this page, add ?viewsource to the URL. -->',
    jsHeader: '/* To view the original source of this page, add ?viewsource to the URL. */',
    configs: {
        development: {
            js: 'source',
            css: 'source',
            images: 'source',
            icons: 'source',
            content: 'source',
        },
        production: {
            js: 'inline',
            css: 'inline',
            images: 'inline',
            icons: 'source',
            content: 'source',
        }
    }
};

var defaultMimeType = 'text/plain';
var htmlMimeType = 'text/html';

var reCrawlers = /googlebot|bingbot|slurp/i;
var reCDNs = /Amazon\sCloudFront/i;

var reRootFiles = /(robots\.txt)|(favicon\.ico)|(apple-touch-icon(.*?)\.png)/;

// *************************************************************************************************

function App(conf) {
    this._assignConfig(conf);

    this.base = addTrailingSlash(this.base);
    this.baseLength = this.base.split("/").length-2;
    this.startupTime = new Date().getTime();

    if (!this.cache) {
        this.cache = new Cache(null, false, true, true);
    }
}

exports.App = App;

App.prototype = {
    get resourceBase() {
        return this.base + this.resources;
    },

    get staticPath() {
        return this.resourceBase + "/static";    
    },

    get jsPath() {
        return this.resourceBase + "/js";    
    },

    get cssPath() {
        return this.resourceBase + "/css";    
    },

    get hasPath() {
        return this.resourceBase + "/has";    
    },

    get polyfillsPath() {
        return this.resourceBase + "/polyfills";    
    },

    get appcachePath() {
        return this.resourceBase + "/cache.manifest";    
    },

    get appjsPrefix() {
        if (!this.reResourcePrefix) {
            return this.reResourcePrefix = new RegExp("^" + this.resources + "(\.\d+)?(:.*?)?$");
        } else {
            return this.reResourcePrefix;
        }
    },

    route: function() {
        function handler(req, res, next) {
            var urlOptions = readOptions(req.query, this);
            if (isCrawler(req)) {
                urlOptions.content = 'inline';
            }

            urlOptions.userAgent = req.header('user-agent');

            this.loadURL(req.url, renderer, urlOptions, function(err, result) {
                if (err) {
                    dandy.logException(err);
                    send(err, true);
                } else {
                    send(result);
                }
            });

            function send(result, isError) {
                if (result.dependencies) {
                    res.dependencies = _.pluck(result.dependencies, 'path');
                }

                var mimeType = result.mimeType || (result.path
                    ? mime.lookup(result.path) || defaultMimeType
                    : htmlMimeType);

                if (isTextMimeType(mimeType)) {
                    mimeType += '; charset=UTF-8'
                }

                res.header('Content-Type', mimeType);

                if (!isError) {
                    var latestTime = findLatestMtime(result.dependencies || []);
                    if (latestTime) {
                        res.header('ETag', latestTime);
                    }

                    if (result.permanent || isCDN(req)) {
                        res.header('Cache-Control', 'public, max-age=31536000');
                    } else {
                        res.header('Cache-Control', 'public, max-age=0');
                    }               
                }

                res.send(result.body, result.error || 200);
            }            
        }

        var renderer = new Renderer(this, this.api, this.apiCache, this.base);

        if (this.cache) {
            this.cache.on('unmonitor', _.bind(function() {
                this.invalidate();
            }, this));
        }

        var cacheMiddleware = this.cache ? cacheware(this.cache) : noware;
        return [prepware(this), uaware(this), cacheMiddleware, _.bind(handler, this)];
    },

    loadURL: function(URL, renderer, options, cb) {
        var u = url.parse(URL);
        var parts = u.pathname.split('/').slice(this.baseLength);
        var m = this.appjsPrefix.exec(parts[1]);
        if (m) {
            if (parts.length == 2) {
                this._loadAppjs(renderer, options, cb);
            } else {
                if (parts[2] == 'static') {
                    this._loadStatic(parts.slice(3).join('/'), renderer, options, cb);
                } else if (parts[2] == 'js') {
                    var userAgent = m[2] ? unescape(m[2].substr(1)) : '';
                    if (userAgent) {
                        options.userAgent = userAgent;
                    }

                    this._loadJS(parts.slice(3).join('/'), renderer, options, cb);
                } else if (parts[2] == 'css') {
                    this._loadCSS(renderer, options, cb);
                } else if (parts[2] == 'cache.manifest') {
                    this._loadAppcache(renderer, options, cb);
                } else if (parts[2] == 'has') {
                    this._loadHas(renderer, options, cb);
                } else if (parts[2] == 'polyfills') {
                    this._loadPolyfills(renderer, options, cb);
                } else {
                    cb("Not found");
                }
            }
        } else if (reRootFiles.exec(parts[1])) {
            var newURL = this.staticPath + '/' + this.packageName + '/' + parts[1];
            this.loadURL(newURL, renderer, options, cb);
        } else {
            this._loadPage(u, renderer, options, cb);
        }
    },

    getContentScript: function(renderer, cb) {
        if (!this.contentScript) {
            this.loadURL(this.jsPath, renderer, {js: 'inline'}, abind(function(err, js) {
                this.contentScript = vm.createScript(js.body, 'document.js');
                cb(0, this.contentScript);
            }, cb, this));
        } else {
            cb(0, this.contentScript);
        }
    },

    invalidate: function() {
        delete this.contentScript;
        delete this.featureDB;
        delete this.dependencies;
    },

    // *********************************************************************************************

    _assignConfig: function(conf) {
        for (var name in defaults) {
            if (name == "configs") {
                var branch = process.env.NODE_ENV == "production"
                             ? defaults.configs.production
                             : defaults.configs.development;
                for (var name in branch) {
                    this[name] = name in conf ? conf[name] : branch[name];
                }
            } else {
                this[name] = name in conf ? conf[name] : defaults[name];
            }
        }        
    },

    _loadPage: function(URL, renderer, options, cb) {
        renderer.renderPage(URL, options, abind(function(err, result) {
            cb(0, {
                body: result.source,
                permanent: options.js == "inline" && options.css == "inline",
                dependencies: result.dependencies,
            });
        }, cb));        
    },

    _loadStatic: function(staticPath, renderer, options, cb) {
        pkg.searchStatic(staticPath, abind(function(err, filePath) {
            if (filePath) {
                fs.lstat(filePath, abind(function(err, stat) {
                    var dependencies = [{path: filePath, mtime: stat.mtime.getTime()}];
                    fs.readFile(filePath, abind(function(err, source) {
                        if (mime.lookup(filePath) == 'text/css') {
                            renderer.compressCSS(source+'', staticPath, options, function(err, source) {
                                cb(0, {path: filePath, body: source, dependencies: dependencies});
                            });
                        } else {
                            cb(0, {path: filePath, body: source, raw: true, dependencies: dependencies});
                        }
                    }, cb, this));                  
                }, cb, this));
            }
        }, cb, this));
    },

    _loadJS: function(modulePath, renderer, options, cb) {
        var scriptPath = this._cleanJSPath(modulePath, renderer);
        var basePath = scriptPath == this.client ? null : this.client;
        pkg.searchScript(scriptPath, basePath, abind(function(err, findings) {
            if (findings) {
                renderer.renderScript(findings.path, findings.name, options, true,
                    abind(function(err, result) {
                        cb(0, {path: findings.path, name: findings.name, body: result.source,
                               dependencies: result.dependencies, permanent: true});
                    }, cb)
                );
            }
        }, cb, this));
    },

    _loadAppjs: function(renderer, options, cb) {
        var scriptName;
        if (options.js == "inline") {
            scriptName = "appjs-inline.js";
        } else if (options.js == "standalone") {
            scriptName = "appjs-standalone.js";            
        } else {
            scriptName = "appjs.js";            
        }
        var appjsPath = path.resolve(__dirname, '..', 'client', scriptName);
        renderer.renderScriptRaw(appjsPath, '', options,
            abind(function(err, result) {
                cb(0, {path: appjsPath, body: result.source, dependencies: result.dependencies});
            }, cb)
        );
    },

    _loadHas: function(renderer, options, cb) {
        renderer.renderHas(options, abind(function(err, result) {
            cb(0, {path: result.path, body: result.source, dependencies: result.dependencies});
        }, cb));
    },

    _loadPolyfills: function(renderer, options, cb) {
        renderer.renderPolyfills(options, abind(function(err, result) {
            cb(0, {path: result.path, body: result.source, dependencies: result.dependencies});
        }, cb));
    },

    _loadCSS: function(renderer, options, cb) {
        renderer.renderStylesheet(options,
            abind(function(err, result) {
                cb(0, {path: result.path, body: result.source, dependencies: result.dependencies,
                       permanent: true});
            }, cb)
        );
    },

    _loadAppcache: function(renderer, options, cb) {
        renderer.renderAppcache(options,
            abind(function(err, result) {
                cb(0, {
                    body: result.source,
                    mimeType: 'text/cache-manifest',
                    permanent: options.js == "inline" && options.css == "inline"
                });
            }, cb)
        );
    },

    _prepare: function(cb) {
        if (!this.dependencies) {
            pkg.traceScript(this.client, false, abind(function(err, deps) {
                this.dependencies = deps;

                async.parallel([
                    ibind(function(next) {
                        hascan.getFeatureDB(_.keys(deps.has), abind(function(err, featureDB) {
                            this.featureDB = featureDB;
                            next(0);
                        }, next, this));                        
                    }, cb, this),

                    ibind(function(next) {
                        if (!this.packageName) {
                            pkg.searchScript(this.client, abind(function(err, findings) {
                                pkg.findPackageInfo(findings.path, abind(function(err, result) {
                                    this.packageName = result.info.name;
                                    next(0);
                                }, cb, this));
                            }, cb, this));
                        } else {
                            next(0);
                        }
                    }, cb, this),

                    ibind(function(next) {
                        if (!this.polyfills) {
                            var polyfillsDirPath = path.join(__dirname, '..', 'client', 'polyfills');
                            this.polyfills = {};
                            fs.readdir(polyfillsDirPath, abind(function(err, names) {
                                if (!err) {
                                    _.each(names, ibind(function(name) {
                                        var feature = path.basename(name, '.js');
                                        this.polyfills[feature] = true;
                                    }, cb, this));
                                }
                                next(0, deps);
                            }, cb, this));
                        } else {
                            next(0);
                        }
                    }, cb, this),
                ], abind(function(err, results) {
                    cb(0, deps);
                }, cb, this));
            }, cb, this));
        } else {
            cb(0, this.dependencies);
        }        
    },

    _cleanJSPath: function(jsPath, renderer) {
        return jsPath
            ? path.join(path.dirname(jsPath), path.basename(jsPath, '.js'))
            : this.client;        
    }
};

// *************************************************************************************************

function prepware(app) {
    return function(req, res, next) {
        app._prepare(function(err, deps) {
            next();
        });
    }
}

function uaware(app) {
    return function(req, res, next) {
        if (req.url == app.base) {
            var featureMap = app.featureDB.getFeatureMap(req.header('user-agent'), true);
            if (!featureMap) {
                forceContentInline();
            } else {
                for (var name in featureMap) {
                    var isRequired = app.dependencies.has[name];
                    if (isRequired && !featureMap[name] && !this.polyfills[name]) {
                        forceContentInline();
                        break;
                    }
                }
            }
        }

        function forceContentInline() {
            var U = url.parse(req.url, true);
            U.query['content'] = 'inline';
            req.url = url.format(U);
            req.query = U.query;
        }
        next();    
    }
}

function noware(req, res, next) {
    next();
}

function readOptions(query, defaults) {
    var viewSource = 'viewsource' in query;
    return {
        js: (viewSource ? 'source' : 'js' in query ? query.js : defaults.js),
        css: (viewSource ? 'source' : 'css' in query ? query.css : defaults.css),
        images: (viewSource ? 'source' : 'images' in query ? query.images : defaults.images),
        content: (viewSource ? 'source' : 'content' in query ? query.content : defaults.content),
        icons: (viewSource ? 'source' : 'icons' in query ? query.icons : defaults.icons),
    };
}

function findLatestMtime(deps) {
    var maxTime = 0;
    _.each(deps, function(dep) {
        if (dep.mtime > maxTime) {
            maxTime = dep.mtime;
        }
    });
    return maxTime;
}

function isCrawler(req) {
    var userAgent = req.headers['user-agent'];
    return !!reCrawlers.exec(userAgent);
}

function isCDN(req) {
    var userAgent = req.headers['user-agent'];
    return !!reCDNs.exec(userAgent);
}

function isTextMimeType(mimeType) {
    return mimeType.indexOf('text/') == 0
        || mimeType == 'application/json'
        || mimeType == 'application/javascript'
        || mimeType == 'application/x-javascript';
}

function addTrailingSlash(s) {
    if (s && s.substr(s.length-1) == '/') {
        return s;
    } else {
        return s + '/';
    }
}

==> app.js-0.0.4.tgz: package/lib/App.js <==

var path = require('path'),
    fs = require('fs'),
    url = require('url'),
    _ = require('underscore'),
    vm = require('vm'),
    async = require('async'),
    assert = require('assert').ok,
    dandy = require('dandy/errors'),
    abind = require('dandy/errors').abind,
    ibind = require('dandy/errors').ibind,
    Cache = require('diskcache').Cache,
    cacheware = require('express-cache'),
    hascan = require('hascan'),
    mime = require('mime'),
    pkg = require('./pkg');
    Renderer = require('./Renderer').Renderer;

// *************************************************************************************************

var defaults = {
    title: '',
    client: null,
    cdn: null,
    api: null,
    apiCache: null,
    cache: null,
    base: "/",
    resources: "app.js",
    language: 'en',
    charset: 'utf8',
    offline: false,
    viewport: null,
    rss: null,
    appcache: null,
    webAppCapable: false,
    statusBarStyle: null,
    inlineImageMaxSize: 1000,
    htmlHeader: '<!-- To view the original source of this page, add ?viewsource to the URL. -->',
    jsHeader: '/* To view the original source of this page, add ?viewsource to the URL. */',
    configs: {
        development: {
            js: 'source',
            css: 'source',
            images: 'source',
            icons: 'source',
            content: 'source',
        },
        production: {
            js: 'inline',
            css: 'inline',
            images: 'inline',
            icons: 'source',
            content: 'source',
        }
    }
};

var defaultMimeType = 'text/plain';
var htmlMimeType = 'text/html';

var reCrawlers = /googlebot|bingbot|slurp/i;
var reCDNs = /Amazon\sCloudFront/i;

var reRootFiles = /(robots\.txt)|(favicon\.ico)|(apple-touch-icon(.*?)\.png)/;

// *************************************************************************************************

function App(conf) {
    this._assignConfig(conf);

    this.base = addTrailingSlash(this.base);
    this.baseLength = this.base.split("/").length-2;
    this.startupTime = new Date().getTime();

    if (!this.cache) {
        this.cache = new Cache(null, false, true, true);
    }
}

exports.App = App;

App.prototype = {
    get resourceBase() {
        return this.base + this.resources;
    },

    get staticPath() {
        return this.resourceBase + "/static";    
    },

    get jsPath() {
        return this.resourceBase + "/js";    
    },

    get cssPath() {
        return this.resourceBase + "/css";    
    },

    get hasPath() {
        return this.resourceBase + "/has";    
    },

    get polyfillsPath() {
        return this.resourceBase + "/polyfills";    
    },

    get appcachePath() {
        return this.resourceBase + "/cache.manifest";    
    },

    get appjsPrefix() {
        if (!this.reResourcePrefix) {
            return this.reResourcePrefix = new RegExp("^" + this.resources + "(\\.\\d+)?(:.*?)?$");
        } else {
            return this.reResourcePrefix;
        }
    },

    route: function() {
        function handler(req, res, next) {
            var urlOptions = readOptions(req.query, this);
            if (isCrawler(req)) {
                urlOptions.content = 'inline';
            }

            urlOptions.userAgent = req.header('user-agent');

            this.loadURL(req.url, renderer, urlOptions, function(err, result) {
                if (err) {
                    dandy.logException(err);
                    send(err, true);
                } else {
                    send(result);
                }
            });

            function send(result, isError) {
                if (result.dependencies) {
                    res.dependencies = _.pluck(result.dependencies, 'path');
                }

                var mimeType = result.mimeType || (result.path
                    ? mime.lookup(result.path) || defaultMimeType
                    : htmlMimeType);

                if (isTextMimeType(mimeType)) {
                    mimeType += '; charset=UTF-8'
                }

                res.header('Content-Type', mimeType);

                if (!isError) {
                    var latestTime = findLatestMtime(result.dependencies || []);
                    if (latestTime) {
                        res.header('ETag', latestTime);
                    }

                    if (result.permanent || isCDN(req)) {
                        res.header('Cache-Control', 'public, max-age=31536000');
                    } else {
                        res.header('Cache-Control', 'public, max-age=0');
                    }               
                }

                res.send(result.body, result.error || 200);
            }            
        }

        var renderer = new Renderer(this, this.api, this.apiCache, this.base);

        if (this.cache) {
            this.cache.on('unmonitor', _.bind(function() {
                this.invalidate();
            }, this));
        }

        var cacheMiddleware = this.cache ? cacheware(this.cache) : noware;
        return [prepware(this), uaware(this), cacheMiddleware, _.bind(handler, this)];
    },

    loadURL: function(URL, renderer, options, cb) {
        var u = url.parse(URL);
        var parts = u.pathname.split('/').slice(this.baseLength);
        var m = this.appjsPrefix.exec(parts[1]);
        if (m) {
            if (parts.length == 2) {
                this._loadAppjs(renderer, options, cb);
            } else {
                if (parts[2] == 'static') {
                    this._loadStatic(parts.slice(3).join('/'), renderer, options, cb);
                } else if (parts[2] == 'js') {
                    var userAgent = m[2] ? unescape(m[2].substr(1)) : '';
                    if (userAgent) {
                        options.userAgent = userAgent;
                    }

                    this._loadJS(parts.slice(3).join('/'), renderer, options, cb);
                } else if (parts[2] == 'css') {
                    this._loadCSS(renderer, options, cb);
                } else if (parts[2] == 'cache.manifest') {
                    this._loadAppcache(renderer, options, cb);
                } else if (parts[2] == 'has') {
                    this._loadHas(renderer, options, cb);
                } else if (parts[2] == 'polyfills') {
                    this._loadPolyfills(renderer, options, cb);
                } else {
                    cb("Not found");
                }
            }
        } else if (reRootFiles.exec(parts[1])) {
            var newURL = this.staticPath + '/' + this.packageName + '/' + parts[1];
            this.loadURL(newURL, renderer, options, cb);
        } else {
            this._loadPage(u, renderer, options, cb);
        }
    },

    getContentScript: function(renderer, cb) {
        if (!this.contentScript) {
            this.loadURL(this.jsPath, renderer, {js: 'inline'}, abind(function(err, js) {
                this.contentScript = vm.createScript(js.body, 'document.js');
                cb(0, this.contentScript);
            }, cb, this));
        } else {
            cb(0, this.contentScript);
        }
    },

    invalidate: function() {
        delete this.contentScript;
        delete this.featureDB;
        delete this.dependencies;
    },

    // *********************************************************************************************

    _assignConfig: function(conf) {
        for (var name in defaults) {
            if (name == "configs") {
                var branch = process.env.NODE_ENV == "production"
                             ? defaults.configs.production
                             : defaults.configs.development;
                for (var name in branch) {
                    this[name] = name in conf ? conf[name] : branch[name];
                }
            } else {
                this[name] = name in conf ? conf[name] : defaults[name];
            }
        }        
    },

    _loadPage: function(URL, renderer, options, cb) {
        renderer.renderPage(URL, options, abind(function(err, result) {
            cb(0, {
                body: result.source,
                permanent: options.js == "inline" && options.css == "inline",
                dependencies: result.dependencies,
            });
        }, cb));        
    },

    _loadStatic: function(staticPath, renderer, options, cb) {
        pkg.searchStatic(staticPath, abind(function(err, filePath) {
            if (filePath) {
                fs.lstat(filePath, abind(function(err, stat) {
                    var dependencies = [{path: filePath, mtime: stat.mtime.getTime()}];
                    fs.readFile(filePath, abind(function(err, source) {
                        if (mime.lookup(filePath) == 'text/css') {
                            renderer.compressCSS(source+'', staticPath, options, function(err, source) {
                                cb(0, {path: filePath, body: source, dependencies: dependencies});
                            });
                        } else {
                            cb(0, {path: filePath, body: source, raw: true, dependencies: dependencies});
                        }
                    }, cb, this));                  
                }, cb, this));
            }
        }, cb, this));
    },

    _loadJS: function(modulePath, renderer, options, cb) {
        var scriptPath = this._cleanJSPath(modulePath, renderer);
        var basePath = scriptPath == this.client ? null : this.client;
        pkg.searchScript(scriptPath, basePath, abind(function(err, findings) {
            if (findings) {
                renderer.renderScript(findings.path, findings.name, options, true,
                    abind(function(err, result) {
                        cb(0, {path: findings.path, name: findings.name, body: result.source,
                               dependencies: result.dependencies, permanent: true});
                    }, cb)
                );
            }
        }, cb, this));
    },

    _loadAppjs: function(renderer, options, cb) {
        var scriptName;
        if (options.js == "inline") {
            scriptName = "appjs-inline.js";
        } else if (options.js == "standalone") {
            scriptName = "appjs-standalone.js";            
        } else {
            scriptName = "appjs.js";            
        }
        var appjsPath = path.resolve(__dirname, '..', 'client', scriptName);
        renderer.renderScriptRaw(appjsPath, '', options,
            abind(function(err, result) {
                cb(0, {path: appjsPath, body: result.source, dependencies: result.dependencies});
            }, cb)
        );
    },

    _loadHas: function(renderer, options, cb) {
        renderer.renderHas(options, abind(function(err, result) {
            cb(0, {path: result.path, body: result.source, dependencies: result.dependencies});
        }, cb));
    },

    _loadPolyfills: function(renderer, options, cb) {
        renderer.renderPolyfills(options, abind(function(err, result) {
            cb(0, {path: result.path, body: result.source, dependencies: result.dependencies});
        }, cb));
    },

    _loadCSS: function(renderer, options, cb) {
        renderer.renderStylesheet(options,
            abind(function(err, result) {
                cb(0, {path: result.path, body: result.source, dependencies: result.dependencies,
                       permanent: true});
            }, cb)
        );
    },

    _loadAppcache: function(renderer, options, cb) {
        renderer.renderAppcache(options,
            abind(function(err, result) {
                cb(0, {
                    body: result.source,
                    mimeType: 'text/cache-manifest',
                    permanent: options.js == "inline" && options.css == "inline"
                });
            }, cb)
        );
    },

    _prepare: function(cb) {
        if (!this.dependencies) {
            pkg.traceScript(this.client, false, abind(function(err, deps) {
                this.dependencies = deps;

                async.parallel([
                    ibind(function(next) {
                        hascan.getFeatureDB(_.keys(deps.has), abind(function(err, featureDB) {
                            this.featureDB = featureDB;
                            next(0);
                        }, next, this));                        
                    }, cb, this),

                    ibind(function(next) {
                        if (!this.packageName) {
                            pkg.searchScript(this.client, abind(function(err, findings) {
                                pkg.findPackageInfo(findings.path, abind(function(err, result) {
                                    this.packageName = result.info.name;
                                    next(0);
                                }, cb, this));
                            }, cb, this));
                        } else {
                            next(0);
                        }
                    }, cb, this),

                    ibind(function(next) {
                        if (!this.polyfills) {
                            var polyfillsDirPath = path.join(__dirname, '..', 'client', 'polyfills');
                            this.polyfills = {};
                            fs.readdir(polyfillsDirPath, abind(function(err, names) {
                                if (!err) {
                                    _.each(names, ibind(function(name) {
                                        var feature = path.basename(name, '.js');
                                        this.polyfills[feature] = true;
                                    }, cb, this));
                                }
                                next(0, deps);
                            }, cb, this));
                        } else {
                            next(0);
                        }
                    }, cb, this),
                ], abind(function(err, results) {
                    cb(0, deps);
                }, cb, this));
            }, cb, this));
        } else {
            cb(0, this.dependencies);
        }        
    },

    _cleanJSPath: function(jsPath, renderer) {
        return jsPath
            ? path.join(path.dirname(jsPath), path.basename(jsPath, '.js'))
            : this.client;        
    }
};

// *************************************************************************************************

function prepware(app) {
    return function(req, res, next) {
        app._prepare(function(err, deps) {
            next();
        });
    }
}

function uaware(app) {
    return function(req, res, next) {
        if (req.url == app.base) {
            var featureMap = app.featureDB.getFeatureMap(req.header('user-agent'), true);
            if (!featureMap) {
                forceContentInline();
            } else {
                for (var name in featureMap) {
                    var isRequired = app.dependencies.has[name];
                    if (isRequired && !featureMap[name] && !this.polyfills[name]) {
                        forceContentInline();
                        break;
                    }
                }
            }
        }

        function forceContentInline() {
            var U = url.parse(req.url, true);
            U.query['content'] = 'inline';
            req.url = url.format(U);
            req.query = U.query;
        }
        next();    
    }
}

function noware(req, res, next) {
    next();
}

function readOptions(query, defaults) {
    var viewSource = 'viewsource' in query;
    return {
        js: (viewSource ? 'source' : 'js' in query ? query.js : defaults.js),
        css: (viewSource ? 'source' : 'css' in query ? query.css : defaults.css),
        images: (viewSource ? 'source' : 'images' in query ? query.images : defaults.images),
        content: (viewSource ? 'source' : 'content' in query ? query.content : defaults.content),
        icons: (viewSource ? 'source' : 'icons' in query ? query.icons : defaults.icons),
    };
}

function findLatestMtime(deps) {
    var maxTime = 0;
    _.each(deps, function(dep) {
        if (dep.mtime > maxTime) {
            maxTime = dep.mtime;
        }
    });
    return maxTime;
}

function isCrawler(req) {
    var userAgent = req.headers['user-agent'];
    return !!reCrawlers.exec(userAgent);
}

function isCDN(req) {
    var userAgent = req.headers['user-agent'];
    return !!reCDNs.exec(userAgent);
}

function isTextMimeType(mimeType) {
    return mimeType.indexOf('text/') == 0
        || mimeType == 'application/json'
        || mimeType == 'application/javascript'
        || mimeType == 'application/x-javascript';
}

function addTrailingSlash(s) {
    if (s && s.substr(s.length-1) == '/') {
        return s;
    } else {
        return s + '/';
    }
}
